'From Squeak 1.2 of June 29, 1997 on 19 October 1998 at 9:05:15 am'!ParseNode subclass: #LanguageCNode
	instanceVariableNames: 'parentNode parser encoder text translation type inputs output code codeIndex '
	classVariableNames: 'Entries '
	poolDictionaries: ''
	category: 'System-Compiler'!

!LanguageCNode methodsFor: 'recording and processing'!
computeTranslation: stream
	"Left justifies each line of code and appends it to the information already in the stream."
	| lines prefixSize firstNonBlank result |
	"Get the lines..."
	lines := code lines.
	"Determine how much we can left justify by ..."
	prefixSize := nil.
	lines do: [:line |
		firstNonBlank := 1.
		[firstNonBlank < line size and: [(line at: firstNonBlank) isSeparator]] whileTrue: [
			firstNonBlank := firstNonBlank + 1].
		firstNonBlank < line size ifTrue: [
			prefixSize := prefixSize isNil ifTrue: [firstNonBlank] ifFalse: [firstNonBlank min: prefixSize]]].
	prefixSize := prefixSize isNil ifTrue: [0] ifFalse: [prefixSize - 1].
	"Left justify the lines."
	lines := lines collect: [:line |
		line size > prefixSize ifTrue: [line suffix: line size - prefixSize] ifFalse: [line]].
	"Eliminate initial and final blank lines."
	[lines notEmpty and: [lines first size = 0]] whileTrue: [lines removeFirst].
	[lines notEmpty and: [lines last size = 0]] whileTrue: [lines removeLast].
	"Append to the existing stream and return the combined result ..."
	lines do: [:line | stream nextPutAll: line; cr].	
	result := stream contents.
	(result notEmpty and: [result first = Character cr]) ifTrue: [result := result suffix: result size - 1].
	^result

	! !

!LanguageCNode methodsFor: 'recording and processing'!
emitReference: reference to: name asType: cType on: stream
	stream cr.
	cType = '?' ifTrue: [
		^stream << 'long ' << name << ' = ' << reference << ';'].
	cType = 'long' ifTrue: [
		^stream << 'long ' << name << ' = asCheckedLong (' << reference << ');'].
	cType = 'pointer' ifTrue: [
		^stream << 'long *' << name << ' = asCheckedPointer (' << reference << ');'].
	cType = 'object' ifTrue: [
		^stream << 'long' << name << ' = ' << reference << ';'].
	cType = 'double' ifTrue: [
		^stream << 'double &' << name << ' = asCheckedDouble (' << reference << ');'].
	self error: 'Unknown c language type ', cType! !

!LanguageCNode methodsFor: 'recording and processing' stamp: 'The Ghost 10/9/98 09:26'!
extractCode
	"Extract the inputs in the format: type1 name1 type2 name2 ..."
	| start end start1 start2 |
	start1 := text indexOfSubCollection: ':in'. 
	start2 := text indexOfSubCollection: ':out'. 
	start := start1 max: start2.
	start = 0 ifTrue: [^false].
	end := text indexOf: $; startingAt: start.
	end = 0 ifTrue: [^false].
	code := text copyFrom: end + 1 to: text size - 1.
	^true! !

!LanguageCNode methodsFor: 'recording and processing' stamp: 'The Ghost 10/9/98 09:20'!
extractInputs
	"Extract the inputs in the format: type1 name1 type2 name2 ..."
	| start end |
	start := text indexOfSubCollection: ':in'. 
	start = 0 ifTrue: [^false].
	end := text indexOf: $; startingAt: start.
	end = 0 ifTrue: [^false].
	inputs := (text copyFrom: start + 3 to: end - 1) findTokens: ',; ', String separators.
	^true! !

!LanguageCNode methodsFor: 'recording and processing' stamp: 'The Ghost 10/9/98 09:47'!
extractOutput
	"Extract the output in the format: type"
	| start end |
	start := text indexOfSubCollection: ':out'. 
	start = 0 ifTrue: [^false].
	end := text indexOf: $; startingAt: start.
	end = 0 ifTrue: [^false].
	output := (text copyFrom: start + 4 to: end - 1) findTokens: ',; ', String separators.
	output isEmpty ifTrue: [output _ 'void'] ifFalse: [output _ output first].
	^true! !

!LanguageCNode methodsFor: 'recording and processing'!
generateCallTranslation
	| stream variable reference |
	stream := WriteStream on: (String new: 255).
	inputs pairsDo: [:cType :name |
		encoder reserveCVariable: name.
		reference := encoder variableReference: name.
		self emitReference: reference to: name asType: cType on: stream.
		self class debugging ifTrue: [
			Transcript cr; show: 'Variable ', name, ' of type ', cType, ' is accessed via ', reference]].
	stream cr.
	translation := self computeTranslation: stream.
	codeIndex := self class register: translation forMethod: parser methodSelector inClass: encoder classEncoding.
	^true! !

!LanguageCNode methodsFor: 'recording and processing' stamp: 'The Ghost 10/9/98 10:06'!
generateModuleTranslation
	translation := text copyFrom: 9 to: text size - 1.! !

!LanguageCNode methodsFor: 'recording and processing'!
parser: aParser text: aString
	parser := aParser. encoder := parser encoder. text := aString.
	^self processText
"
	(text prefix: 8) = '{:module' ifTrue: [^true].
	(text prefix: 4) = '{:in' ifTrue: [^true].
	(text prefix: 5) = '{:out' ifTrue: [^true].
	^false
"! !

!LanguageCNode methodsFor: 'recording and processing'!
processCall
	type := #call.
	self extractInputs ifFalse: [^false].
	self extractOutput ifFalse: [^false].
	self extractCode ifFalse: [^false].
	self generateCallTranslation.
	self class debugging ifTrue: [
		
Transcript cr; show: 'Back from generateCallTranslation'.
		Transcript cr; show: 'Inputs ', inputs asArray printString.
		Transcript cr; show: 'Output ', output printString.
		Transcript cr; show: 'Code ', code.
		Transcript cr; show: 'Translation '; cr; show: translation].
	^true! !

!LanguageCNode methodsFor: 'recording and processing'!
processModule
	type := #module.
	self generateModuleTranslation.
	self class debugging ifTrue: [
		Transcript cr; show: 'Module is ', translation].
	^true! !

!LanguageCNode methodsFor: 'recording and processing' stamp: 'The Ghost 10/9/98 08:48'!
processText
	(text prefix: 8) = '{:module' ifTrue: [^self processModule].
	(text prefix: 4) = '{:in' ifTrue: [^self processCall].
	(text prefix: 5) = '{:out' ifTrue: [^self processCall].
	^false! !


!LanguageCNode methodsFor: 'tree generation'!
insertIfFailAbove
	| selfVariable failSelector failMessage block ifFailSelector ifFailMessage |
	selfVariable := encoder encodeVariable: 'self'.
	failSelector := encoder encodeSelector: #fail.
"SelectorNode new key: #fail index: 0 type: SendType."
	failMessage := MessageNode new
		receiver: selfVariable selector: failSelector arguments: #() precedence: 1.
	block := BlockNode new statements: (Array with: failMessage) returns: false.
	ifFailSelector := encoder encodeSelector: #ifFail:.
"SelectorNode new key: #ifFail: index: 0 type: SendType."
	ifFailMessage := MessageNode new
		receiver: self shallowCopy selector: ifFailSelector arguments: (Array with: block) precedence: 3.
	self become: ifFailMessage
	 ! !


!LanguageCNode methodsFor: 'code generation'!
emitForValue: stack on: strm

self halt: 'Should not be reached'.

	((parentNode isKindOf: MessageNode) and: [parentNode selector key == #ifFail:]) ifTrue: [

		Transcript cr; show: 'Yes, parent is ifFail:'] ifFalse: [

		Transcript cr; show: 'No, parent is not ifFail:'. Transcript cr; show: 'It''s ', parentNode class printString, ' ', parentNode printString, ' selector = ', parentNode selector printString].

	strm nextPut: ExternalCall; nextPut: codeIndex >> 8; nextPut: (codeIndex bitAnd: 16rFF).

	stack push: 1! !

!LanguageCNode methodsFor: 'code generation'!
emitIfFail: ifFailMessage forValue: stack on: stream

	| ifFailBlockSize ifFailExpression start end |

	"CHANGED MY MIND: First, process the languageCNode."
	"self processText."


	"Next, generate the external call instruction with the caseIndex."

	stream nextPut: ExternalCall; nextPut: codeIndex >> 8; nextPut: (codeIndex bitAnd: 16rFF).
"stack push: 1."



	"Next the size of the ifFail: code."

	ifFailExpression := ifFailMessage arguments at: 1.

	ifFailBlockSize _ ifFailExpression sizeForEvaluatedValue: encoder.
	self class debugging ifTrue: [
		Transcript cr; show: 'Block size is said to be ', ifFailBlockSize printString, ' bytes.'].

	stream nextPut: ifFailBlockSize >> 8; nextPut: (ifFailBlockSize bitAnd: 16rFF).
start := stream position.

	ifFailExpression emitForEvaluatedValue: stack on: stream.
"stack pop: 1."
end := stream position.
	self class debugging ifTrue: [
		
Transcript cr; show: 'Block size turns out to be ', (end - start) printString, ' = (', end printString, ' - ', start printString, ') bytes.'.
		Transcript cr; show: 'Literals = ', encoder literals printString]! !

!LanguageCNode methodsFor: 'code generation'!
sizeForValue: anEncoder

	"ExternalCall byte1Index byte2Index byte1PastFailBlockOffset byte2PastFailBlockOffset"

	^5! !


!LanguageCNode methodsFor: 'language C scanning'!
addParent: parent

	parentNode := parent.

	((parentNode isKindOf: MessageNode) and: [parentNode selector key == #ifFail:]) ifFalse: [

		^self insertIfFailAbove]  ! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

LanguageCNode class
	instanceVariableNames: ''!

!LanguageCNode class methodsFor: 'initializing' stamp: 'The Ghost 10/17/98 17:15'!
debugging
	^false! !

!LanguageCNode class methodsFor: 'initializing'!
initialize
	"LanguageCNode initialize"
	Entries _ OrderedCollection new.! !

!LanguageCNode class methodsFor: 'initializing'!
setup
	"LanguageCNode setup"
	Entries isNil ifTrue: [Entries _ ObjectFiler loadFromFileName: 'externals']! !

!LanguageCNode class methodsFor: 'initializing'!
wrapup
	"LanguageCNode wrapup"
	Entries isNil ifFalse: [ObjectFiler dump: Entries onFileName: 'externals'].
	Entries := nil! !


!LanguageCNode class methodsFor: 'registering/deregistering'!
deregisterMethod: selector inClass: aClass
	"Eliminates all old code."
	| count |
	Entries isNil ifTrue: [self setup].
	count := 0.
	Entries groupsDo: [:className :entrySelector :codeCollection |
		count := count + 1.
		(selector == entrySelector and: [className = aClass name]) ifTrue: [
			Entries removeAt: count.
			^self]].
	"Nothing there."! !

!LanguageCNode class methodsFor: 'registering/deregistering' stamp: 'The Ghost 10/11/98 10:45'!
nextUnusedEntry
	| indices |
	indices := SortedCollection new.
	Entries groupsDo: [:className :entrySelector :codeCollection |
		codeCollection groupsDo: [:index :code |
			indices add: index]].
	indices isEmpty ifTrue: [^1].
	indices first ~= 1 ifTrue: [^1].
	indices indexedDo: [:element :index |
		element ~= index ifTrue: [^index]].
	^Entries size + 1! !

!LanguageCNode class methodsFor: 'registering/deregistering'!
register: code forMethod: selector inClass: aClass
	"Records the code and returns the index of the code."
	| index |
	Entries isNil ifTrue: [self setup].
	index := self nextUnusedEntry.
	Entries groupsDo: [:className :entrySelector :codeCollection |
		(selector == entrySelector and: [className = aClass name]) ifTrue: [
			codeCollection add: (Array with: index with: code).
			^index]].
	Entries add: (Array with: aClass name with: selector with: (OrderedCollection with: (Array with: index with: code))).
	^index! !


!LanguageCNode class methodsFor: 'file generation'!
generateExternalFile
	"LanguageCNode generateExternalFile"
	"Create a file from the entries."
	| all path stream count |
	Entries isNil ifTrue: [self setup].
	all := SortedCollection sortBlock: [:array1 :array2 | array1 first <= array2 first].
	Entries groupsDo: [:className :entrySelector :codeCollection |
		codeCollection groupsDo: [:index :code |
			all add: (Array with: index with: className with: entrySelector with: code)]].


	path := File fullPathName: 'externals.c'.

	stream := File newFile: path.
	stream line. 
	stream line; nextPutAll: 'inline long external (long which) {'.
	stream line; tab; nextPutAll: 'switch (which) {'.
	count := -1.
	all groupsDo: [:index :className :selector :code |
		count := count + 1.
		count < index 
			ifTrue: [
				(stream line; tab; tab) << 'case '.
				stream << count.
				[count := count + 1. count < index] whileTrue: [stream << ', ' << count].
				stream << ': //Unused.'.
				stream line; tab; tab; tab. stream << 'fail (); return 0;'.
				stream line; tab; tab; tab. stream << 'break;'].
		(stream line; tab; tab) << 'case ' << index << ': {//' << className << '>>' << selector << '.'.
		code linesDo: [:line | (stream line; tab; tab; tab) << line].
		(stream line; tab; tab; tab) << 'break;}'].

	stream line; tab; tab; nextPutAll: 'default:'.
	stream line; tab; tab; tab; nextPutAll:
 'error ("Unknown external routine");'.
	stream line; tab; tab; tab; nextPutAll:
 'return 0;'.
	stream line; tab; nextPutAll:
	'}'.
	stream line; nextPutAll: '}'.
	stream close! !

!LanguageCNode class methodsFor: 'file generation'!
generateTest1
	"LanguageCNode generateTest1"
	| path stream |
	path := File fullPathName: 'test1.c'.
	stream := File oldFile: path.
	stream nextPutAll: '
a
b'.
	stream close! !

!LanguageCNode class methodsFor: 'file generation'!
recallTest1
	"LanguageCNode recallTest1"
	| path stream c |
	path := File fullPathName: 'test1.c'.
	stream := File oldFile: path.
	Transcript cr; show: 'Test1.c'.
	[stream atEnd] whileFalse: [
		c := stream next. Transcript cr; show: c; show: ' '; print: c asInteger].
	stream close! !

!LanguageCNode class methodsFor: 'file generation' stamp: 'The Ghost 10/12/98 09:20'!
recallTest2
	"LanguageCNode recallTest2"
	| path stream c |
	path := File fullPathName: 'test2.c'.
	stream := File oldFile: path.
	Transcript cr; show: 'Test2.c'.
	[stream atEnd] whileFalse: [
		c := stream next. Transcript cr; show: c; show: ' '; print: c asInteger].
	stream close! !


LanguageCNode initialize!
